"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.hasDarkMode = exports.resolveThemeExtension = exports.normalizeTheme = exports.replaceThemeRoot = exports.getThemeFilename = exports.replaceTheme = exports.parseThemeKey = void 0;
const path_1 = __importDefault(require("path"));
const fs_1 = __importDefault(require("fs"));
const deepmerge_1 = __importDefault(require("deepmerge"));
const THEME_USAGE_REGEX = /@theme\s+\$?([a-zA-Z-_0-9.]+)/;
const ALT_THEME_USAGE_REGEX = /theme\(\s*['"]([a-zA-Z-_0-9.]+)['"]\s*\)/;
/** Get the theme variable name from a string */
const parseThemeKey = (value) => {
    let key = value.match(THEME_USAGE_REGEX);
    if (key) {
        return key[1];
    }
    key = value.match(ALT_THEME_USAGE_REGEX);
    if (key) {
        return key[1];
    }
    return '';
};
exports.parseThemeKey = parseThemeKey;
/** Replace a theme variable reference with a value */
const replaceTheme = (value, replace) => {
    if (value.match(THEME_USAGE_REGEX)) {
        return value.replace(THEME_USAGE_REGEX, replace);
    }
    return value.replace(ALT_THEME_USAGE_REGEX, replace);
};
exports.replaceTheme = replaceTheme;
/** Get the location of the theme file */
function getThemeFilename(cssFile) {
    let themePath = path_1.default.join(path_1.default.dirname(cssFile), 'theme.ts');
    if (!fs_1.default.existsSync(themePath)) {
        themePath = path_1.default.join(path_1.default.dirname(cssFile), 'theme.js');
    }
    return themePath;
}
exports.getThemeFilename = getThemeFilename;
/** Remove :theme-root usage from a selector */
const replaceThemeRoot = (selector) => selector.replace(/:theme-root\((\S+)\)/g, '$1').replace(/:theme-root/g, '');
exports.replaceThemeRoot = replaceThemeRoot;
/** Make a SimpleTheme into a LightDarkTheme */
const normalizeTheme = (config) => {
    return Object.assign({}, ...Object.entries(config).map(([theme, themeConfig]) => {
        if ('light' in themeConfig && 'dark' in themeConfig) {
            return { [theme]: themeConfig };
        }
        if (themeConfig.extends) {
            const configWithoutExtends = Object.assign({}, themeConfig);
            delete configWithoutExtends.extends;
            return {
                [theme]: {
                    extends: themeConfig.extends,
                    light: configWithoutExtends,
                    dark: {},
                },
            };
        }
        return { [theme]: { light: themeConfig, dark: {} } };
    }));
};
exports.normalizeTheme = normalizeTheme;
/** Resolve any "extends" fields for a theme */
const resolveThemeExtension = (config) => {
    const checkExtendSelf = (theme, extendsTheme) => {
        if (extendsTheme === theme) {
            throw new Error(`A theme cannot extend itself! '${theme}' extends '${extendsTheme}'`);
        }
    };
    const checkThemeExists = (extendsTheme) => {
        if (!config[extendsTheme]) {
            throw new Error(`Theme to extend from not found! '${extendsTheme}'`);
        }
    };
    const checkCycles = (theme, colorScheme) => {
        const chain = [theme];
        let currentTheme = colorScheme
            ? config[theme][colorScheme].extends
            : config[theme].extends;
        while (currentTheme) {
            if (chain.includes(currentTheme)) {
                chain.push(currentTheme);
                throw new Error(`Circular theme extension found! ${chain
                    .map((i) => `'${i}'`)
                    .join(' => ')}`);
            }
            chain.push(currentTheme);
            currentTheme = colorScheme
                ? config[currentTheme][colorScheme].extends
                : config[currentTheme].extends;
        }
    };
    const resolveSubTheme = (theme) => {
        const subConfig = Object.assign({}, config);
        delete subConfig[theme];
        Object.keys(subConfig).forEach((t) => {
            if (subConfig[t].extends === theme ||
                subConfig[t].light.extends === theme ||
                subConfig[t].dark.extends === theme) {
                delete subConfig[t];
            }
        });
        exports.resolveThemeExtension(subConfig);
    };
    const resolveColorSchemeTheme = (themeConfig, theme, colorScheme) => {
        const extendsTheme = themeConfig[colorScheme].extends;
        let extras = {};
        if (extendsTheme) {
            checkThemeExists(extendsTheme);
            checkExtendSelf(theme, extendsTheme);
            checkCycles(theme, colorScheme);
            if (config[extendsTheme][colorScheme].extends) {
                resolveSubTheme(theme);
            }
            extras = config[extendsTheme][colorScheme];
            delete themeConfig[colorScheme].extends;
        }
        return extras;
    };
    Object.entries(config).forEach(([theme, themeConfig]) => {
        let lightExtras = {};
        let darkExtras = {};
        if (themeConfig.extends) {
            checkThemeExists(themeConfig.extends);
            checkExtendSelf(theme, themeConfig.extends);
            checkCycles(theme);
            if (config[themeConfig.extends]) {
                resolveSubTheme(theme);
            }
            const newConfig = deepmerge_1.default(config[themeConfig.extends], themeConfig);
            delete themeConfig.extends;
            themeConfig.light = newConfig.light;
            themeConfig.dark = newConfig.dark;
        }
        if (themeConfig.light.extends) {
            lightExtras = resolveColorSchemeTheme(themeConfig, theme, 'light');
        }
        if (themeConfig.dark.extends) {
            darkExtras = resolveColorSchemeTheme(themeConfig, theme, 'dark');
        }
        themeConfig.light = Object.assign(Object.assign({}, lightExtras), themeConfig.light);
        themeConfig.dark = Object.assign(Object.assign({}, darkExtras), themeConfig.dark);
    });
    return config;
};
exports.resolveThemeExtension = resolveThemeExtension;
/** Determine if a theme has dark mode enabled */
const hasDarkMode = (theme) => Boolean(Object.keys(theme.dark).length > 0 && Object.keys(theme.light).length > 0);
exports.hasDarkMode = hasDarkMode;
//# sourceMappingURL=index.js.map