"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.legacyTheme = void 0;
const postcss_1 = __importDefault(require("postcss"));
const debug_1 = __importDefault(require("debug"));
const dlv_1 = __importDefault(require("dlv"));
const common_1 = require("../common");
const log = debug_1.default('postcss-themed');
/** Find all the theme variables in a CSS value and replace them with the configured theme values */
const replaceThemeVariables = (config, theme, decl, colorScheme = 'light', defaultTheme = 'default') => {
    const hasMultiple = (decl.value.match(/@theme/g) || decl.value.match(/theme\(['"]/g) || [])
        .length > 1;
    let themeKey = common_1.parseThemeKey(decl.value);
    // Found a theme reference
    while (themeKey) {
        // Check for issues with theme
        try {
            const themeDefault = dlv_1.default(config[defaultTheme][colorScheme], themeKey);
            const newValue = dlv_1.default(config[theme][colorScheme], themeKey);
            decl.value = common_1.replaceTheme(decl.value, hasMultiple ? newValue || themeDefault : newValue);
            if (decl.value === 'undefined') {
                decl.remove();
            }
        }
        catch (error) {
            log(error);
            throw decl.error(`Theme '${theme}' does not contain key '${themeKey}'`, {
                plugin: 'postcss-themed',
            });
        }
        themeKey = common_1.parseThemeKey(decl.value);
    }
};
/** Apply a transformation to a selector */
const applyToSelectors = (selector, fn) => {
    return selector.replace(/\n/gm, '').split(',').map(fn).join(',');
};
/** Create a new rule by inject injecting theme vars into a class with theme usage */
const createNewRule = (componentConfig, rule, themedDeclarations, originalSelector, defaultTheme) => (theme, colorScheme) => {
    if (theme === defaultTheme && colorScheme === 'light') {
        return;
    }
    if (Object.keys(componentConfig[theme][colorScheme]).length === 0) {
        return;
    }
    const themeClass = (colorScheme !== 'dark' && `.${theme}`) ||
        (theme === defaultTheme && `.${colorScheme}`) ||
        `.${theme}.${colorScheme}`;
    let newSelector = applyToSelectors(originalSelector, (s) => `${themeClass} ${s}`);
    if (originalSelector.includes(':theme-root')) {
        rule.selector = common_1.replaceThemeRoot(rule.selector);
        if (rule.selector === '*') {
            newSelector = applyToSelectors(rule.selector, (s) => `${s}${themeClass}`);
        }
        else {
            newSelector = applyToSelectors(rule.selector, (s) => `${themeClass}${s}`);
        }
    }
    if (themedDeclarations.length > 0) {
        // Add theme to selector, clone to retain source maps
        const newRule = rule.clone({
            selector: newSelector,
        });
        newRule.removeAll();
        // Only add themed declarations to override
        for (const property of themedDeclarations) {
            const declaration = postcss_1.default.decl(property);
            replaceThemeVariables(componentConfig, theme, declaration, colorScheme, defaultTheme);
            if (declaration.value !== 'undefined') {
                newRule.append(declaration);
            }
        }
        return newRule;
    }
};
/** Create theme override rule for every theme */
const createNewRules = (componentConfig, rule, themedDeclarations, defaultTheme) => {
    // Need to remember original selector because we overwrite rule.selector
    // once :theme-root is found. If we don't remember the original value then
    // multiple themes break
    const originalSelector = rule.selector;
    const themes = Object.keys(componentConfig);
    const rules = [];
    // Create new rules for theme overrides
    for (const themeKey of themes) {
        const theme = componentConfig[themeKey];
        const themeRule = createNewRule(componentConfig, rule, themedDeclarations, originalSelector, defaultTheme);
        for (const colorScheme in theme) {
            const newRule = themeRule(themeKey, colorScheme);
            if (newRule) {
                rules.push(newRule);
            }
        }
    }
    return rules;
};
/** Accomplish theming by creating new classes to override theme values  */
const legacyTheme = (root, componentConfig, options) => {
    const { defaultTheme = 'default', forceSingleTheme = undefined, forceEmptyThemeSelectors, } = options;
    let newRules = [];
    root.walkRules((rule) => {
        const themedDeclarations = [];
        // Walk each declaration and find themed values
        rule.walkDecls((decl) => {
            const { value } = decl;
            if (common_1.parseThemeKey(value)) {
                themedDeclarations.push(decl.clone());
                // Replace defaults in original CSS rule
                replaceThemeVariables(componentConfig, defaultTheme, decl, 'light', defaultTheme);
            }
        });
        let createNewThemeRules;
        if (forceSingleTheme) {
            createNewThemeRules = [];
        }
        else {
            createNewThemeRules = createNewRules(componentConfig, rule, themedDeclarations, defaultTheme);
        }
        newRules = [...newRules, ...createNewThemeRules];
    });
    if (forceEmptyThemeSelectors) {
        const themes = Object.keys(componentConfig);
        const extra = new Set();
        for (const themeKey of themes) {
            const theme = componentConfig[themeKey];
            extra.add(themeKey);
            for (const colorScheme in theme) {
                extra.add(colorScheme);
            }
        }
        extra.forEach((selector) => newRules.push(postcss_1.default.rule({ selector: `.${selector}` })));
    }
    newRules.forEach((r) => {
        if (forceEmptyThemeSelectors || (r.nodes && r.nodes.length > 0)) {
            root.append(r);
        }
    });
};
exports.legacyTheme = legacyTheme;
//# sourceMappingURL=index.js.map